from django.dispatch import Signal


# An event that will be fired prior to an email being sent
pre_send = Signal(providing_args=['email', 'event', 'context', 'message'])

# An event that will be fired if an exception occurs when trying to send an email
email_exception = Signal(providing_args=['email', 'exception'])



# Generated by Django 2.2.10 on 2020-02-10 18:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('entity_emailer', '0002_auto_20170919_1653'),
    ]

    operations = [
        migrations.AddField(
            model_name='email',
            name='exception',
            field=models.TextField(default=None, null=True),
        ),
    ]



import sys
import traceback

from datetime import datetime

from django.core import mail
from entity_event import context_loader

from entity_emailer.models import Email
from entity_emailer.signals import pre_send, email_exception

from entity_emailer.utils import get_medium, get_from_email_address, get_subscribed_email_addresses, \
    create_email_message, extract_email_subject_from_html_content


class EntityEmailerInterface(object):
    """
    An api interface to do things within entity emailer
    """

    @staticmethod
    def send_unsent_scheduled_emails():
        """
        Send out any scheduled emails that are unsent
        """

        # Get the emails that we need to send
        current_time = datetime.utcnow()
        email_medium = get_medium()
        to_send = Email.objects.filter(
            scheduled__lte=current_time,
            sent__isnull=True
        ).select_related(
            'event__source'
        ).prefetch_related(
            'recipients'
        ).order_by(
            'scheduled',
            'id'
        )

        # Fetch the contexts of every event so that they may be rendered
        context_loader.load_contexts_and_renderers([e.event for e in to_send], [email_medium])

        # Keep track of what emails we will be sending
        emails = []

        # Loop over each email and generate the recipients, and message
        # and handle any exceptions that may occur
        for email in to_send:
            # Compute what email addresses we actually want to send this email to
            to_email_addresses = get_subscribed_email_addresses(email)

            # If there are no recipients we can just skip rendering
            if not to_email_addresses:
                continue

            # If any exceptions occur we will catch the exception and store it as a reference
            # As well as fire off a signal with the error and mark the email as sent and errored
            try:
                # Render the email
                text_message, html_message = email.render(email_medium)

                # Create the email
                message = create_email_message(
                    to_emails=to_email_addresses,
                    from_email=email.from_address or get_from_email_address(),
                    subject=email.subject or extract_email_subject_from_html_content(html_message),
                    text=text_message,
                    html=html_message,
                )

                # Fire the pre send signal
                pre_send.send(
                    sender=sys.intern(email.event.source.name),
                    email=email,
                    event=email.event,
                    context=email.event.context,
                    message=message,
                )

                # Add the email to the list of emails that need to be sent
                emails.append(message)
            except Exception as e:
                # Save the exception on the model
                email.exception = traceback.format_exc()
                email.save(update_fields=['exception'])

                # Fire the email exception event
                email_exception.send(
                    sender=Email,
                    email=email,
                    exception=e
                )

        # Send all the emails that were generated properly
        connection = mail.get_connection()
        connection.send_messages(emails)

        # Update the emails as sent
        to_send.update(sent=current_time)

    @staticmethod
    def convert_events_to_emails():
        """
        Converts unseen events to emails and marks them as seen.
        """

        # Get the email medium
        email_medium = get_medium()

        # Get the default from email
        default_from_email = get_from_email_address()

        # Find any unseen events and create unsent email objects
        for event, targets in email_medium.events_targets(seen=False, mark_seen=True):

            # Check the event's context for a from_address, otherwise fallback to default
            from_address = event.context.get('from_address') or default_from_email

            # Create the emails
            Email.objects.create_email(event=event, from_address=from_address, recipients=targets)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-09-19 16:53
from __future__ import unicode_literals

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('entity_emailer', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='view_uid',
            field=models.UUIDField(default=uuid.uuid4, editable=False),
        ),
    ]



from bs4 import BeautifulSoup
from django.conf import settings
from django.core import mail
from entity_event.models import Medium, Source


constants = {
    'default_medium_name': 'email',
    'default_admin_source_name': 'admin',
}


def get_medium():
    """Get the medium object that the emailer associates with itself.
    """
    email_medium_name = getattr(
        settings, 'ENTITY_EMAILER_MEDIUM_NAME', constants['default_medium_name']
    )
    email_medium = Medium.objects.get(name=email_medium_name)
    return email_medium


def get_admin_source():
    """Get the source object for emails sent from the admin site.
    """
    admin_source_name = getattr(
        settings, 'ENTITY_EMAILER_ADMIN_SOURCE_NAME', constants['default_admin_source_name']
    )
    admin_source = Source.objects.get(name=admin_source_name)
    return admin_source


def get_from_email_address():
    """
    Get a 'from' address based on the django settings.
    """
    return getattr(settings, 'ENTITY_EMAILER_FROM_EMAIL', settings.DEFAULT_FROM_EMAIL)


def get_subscribed_email_addresses(email):
    """
    Given the email recipients, get the email address from the entity metadata.

    The email field is determined in the settings by the ENTITY_EMAILER_EMAIL_KEY field.

    If the user wishes to exclude certain entities from receiving emails, they can define
    which field in the entity metadata to use with the EXCLUDE_ENTITY_EMAILER_KEY field.
    """

    # Get the key to use to find the email address
    email_key = getattr(settings, 'ENTITY_EMAILER_EMAIL_KEY', 'email')

    # Get the exclude key
    exclude_entity_key = getattr(settings, 'ENTITY_EMAILER_EXCLUDE_KEY', None)

    # Get the email addresses from the recipient entities
    email_addresses = []
    for entity in email.recipients.all():
        # Get the email address out of the entity meta data
        email_address = entity.entity_meta.get(email_key, None)

        # Make sure the email address exists and is not an empty string
        if email_address is not None and len(email_address):
            # If the exclude entity key is not set, or is set but the value is not none
            if not exclude_entity_key or entity.entity_meta.get(exclude_entity_key):
                email_addresses.append(email_address)

    # Return the email addresses
    return email_addresses


def create_email_message(to_emails, from_email, subject, text, html):
    """
    Create the appropriate plaintext or html email object.

    Returns:

       email - an instance of either `django.core.mail.EmailMessage` or
       `django.core.mail.EmailMulitiAlternatives` based on whether or
       not `html_message` is empty.
    """
    if not html:
        email = mail.EmailMessage(
            subject=subject,
            body=text,
            to=to_emails,
            from_email=from_email,
        )
    else:
        email = mail.EmailMultiAlternatives(
            subject=subject,
            body=text,
            to=to_emails,
            from_email=from_email,
        )
        email.attach_alternative(html, 'text/html')
    return email


def extract_email_subject_from_html_content(email_content):
    """
    This function extracts an email subject from the rendered html email context.
    It first tries to find a title block inside of a head block. If that exists,
    the title is used as the subject of the email. If it does not exist,
    the first 40 characters of the email are used as the subject. In the latter
    case, it is assumed that html tags are not actually present in the html content.
    """
    soup = BeautifulSoup(email_content, 'html.parser')
    subject = soup.title.string.strip() if soup.title else None
    if not subject:
        subject = email_content.split('\n')[0].strip()[:40]
        if len(subject) == 40:
            subject = u'{}...'.format(subject)

    return subject



from datetime import datetime

from django.test import TestCase
from django_dynamic_fixture import G
from entity.models import Entity
from entity_event.models import Event
from freezegun import freeze_time

from entity_emailer.models import Email


class EmailManagerCreateEmailTest(TestCase):
    @freeze_time('2013-2-3')
    def test_w_recipients_scheduled_time(self):
        e1 = G(Entity)
        e2 = G(Entity)
        event = G(Event, context={'hi': 'hi'})
        e = Email.objects.create_email(
            scheduled=datetime(2013, 4, 5), recipients=[e1, e2],
            subject='hi', from_address='hi@hi.com', event=event, uid='hi')
        self.assertEqual(e.scheduled, datetime(2013, 4, 5))
        self.assertEqual(set(e.recipients.all()), set([e1, e2]))
        self.assertEqual(e.subject, 'hi')
        self.assertEqual(e.from_address, 'hi@hi.com')
        self.assertEqual(e.event.context, {'hi': 'hi'})
        self.assertEqual(e.uid, 'hi')

    @freeze_time('2013-2-3')
    def test_w_recipients_no_scheduled_time(self):
        e1 = G(Entity)
        e2 = G(Entity)
        event = G(Event, context={'hi': 'hi'})
        e = Email.objects.create_email(
            recipients=[e1, e2], subject='hi', from_address='hi@hi.com',
            event=event, uid='hi')
        self.assertEqual(e.scheduled, datetime(2013, 2, 3))
        self.assertEqual(set(e.recipients.all()), set([e1, e2]))
        self.assertEqual(e.subject, 'hi')
        self.assertEqual(e.from_address, 'hi@hi.com')
        self.assertEqual(e.event.context, {'hi': 'hi'})
        self.assertEqual(e.uid, 'hi')

    @freeze_time('2013-2-3')
    def test_w_recipients_no_scheduled_time_no_recipients_no_uid(self):
        event = G(Event, context={'hi': 'hi'})
        e = Email.objects.create_email(
            subject='hi', from_address='hi@hi.com', event=event)
        self.assertEqual(e.scheduled, datetime(2013, 2, 3))
        self.assertEqual(list(e.recipients.all()), [])
        self.assertEqual(e.subject, 'hi')
        self.assertEqual(e.from_address, 'hi@hi.com')
        self.assertEqual(e.event.context, {'hi': 'hi'})
        self.assertIsNone(e.uid)



from django.urls import reverse
from django.test import TestCase
from django_dynamic_fixture import G
from entity.models import Entity
from entity_event.models import Medium, RenderingStyle, ContextRenderer, Source, Event
import six

from entity_emailer.tests.utils import g_email


class EmailViewTest(TestCase):
    def setUp(self):
        self.rendering_style = G(RenderingStyle, name='email')
        G(Medium, name='email', rendering_style=self.rendering_style)
        self.source = G(Source)

    def test_html_path(self):
        G(
            ContextRenderer, source=self.source, html_template_path='hi_template.html',
            rendering_style=self.rendering_style, context_hints={
                'entity': {
                    'app_name': 'entity',
                    'model_name': 'Entity',
                }
            })
        person = G(Entity, display_name='Swansonbot')
        event = G(Event, context={'entity': person.id}, source=self.source)
        email = g_email(event=event)

        url = reverse('entity_emailer.email', args=[email.view_uid])
        response = self.client.get(url)
        content = response.content
        if six.PY3:  # pragma: no cover
            content = content.decode('utf8')

        self.assertEqual(content, '<html>Hi Swansonbot</html>')

    def test_text_path(self):
        G(
            ContextRenderer, source=self.source, html_template_path='hi_template.txt',
            rendering_style=self.rendering_style, context_hints={
                'entity': {
                    'app_name': 'entity',
                    'model_name': 'Entity',
                }
            })
        person = G(Entity, display_name='Swansonbot')
        event = G(Event, context={'entity': person.id}, source=self.source)
        email = g_email(event=event)
        url = reverse('entity_emailer.email', args=[email.view_uid])
        response = self.client.get(url)
        content = response.content
        if six.PY3:  # pragma: no cover
            content = content.decode('utf8')
        self.assertEqual(content, 'Hi Swansonbot')

    def test_text_and_html_path(self):
        G(
            ContextRenderer, source=self.source,
            html_template_path='hi_template.html', text_template_path='hi_template.txt',
            rendering_style=self.rendering_style, context_hints={
                'entity': {
                    'app_name': 'entity',
                    'model_name': 'Entity',
                }
            })

        person = G(Entity, display_name='Swansonbot')
        event = G(Event, context={'entity': person.id}, source=self.source)
        email = g_email(context={'entity': person.id})
        email = g_email(event=event)

        url = reverse('entity_emailer.email', args=[email.view_uid])
        response = self.client.get(url)
        content = response.content
        if six.PY3:  # pragma: no cover
            content = content.decode('utf8')

        self.assertEqual(content, '<html>Hi Swansonbot</html>')



from django_dynamic_fixture import G, F
import uuid

from entity_emailer.models import Email


def g_email(**kwargs):
    view_uid = kwargs.pop('view_uid', uuid.uuid4())
    context = kwargs.pop('context', None)
    if context is not None:
        return G(Email, view_uid=view_uid, event=F(context=context), **kwargs)
    else:
        return G(Email, view_uid=view_uid, **kwargs)



from os import path

import re
from setuptools import setup, find_packages

# import multiprocessing to avoid this bug (http://bugs.python.org/issue15881#msg170215)
import multiprocessing
assert multiprocessing


def get_version():
    """
    Extracts the version number from the version.py file.
    """
    VERSION_FILE = 'entity_emailer/version.py'
    mo = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', open(VERSION_FILE, 'rt').read(), re.M)
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError('Unable to find version string in {0}.'.format(VERSION_FILE))


def get_requirements(requirements_file):
    """
    Gets a list of requirements from requirements.txt.
    """
    with open(path.join(path.dirname(__file__), 'requirements', requirements_file)) as requirements_file:
        requirements = requirements_file.readlines()

    requirements = [r.strip() for r in requirements if r.strip()]

    return [
        r for r in requirements
        if not r.startswith('#') and not r.startswith('-')
    ]

setup(
    name='django-entity-emailer',
    version=get_version(),
    description='Make emailing users easy and entity-based.',
    long_description=open('README.rst').read(),
    url='https://github.com/ambitioninc/django-entity-emailer',
    author='Erik Swanson',
    author_email='opensource@ambition.com',
    keywords='',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
    ],
    license='MIT',
    install_requires=get_requirements('requirements.txt'),
    tests_require=get_requirements('requirements-testing.txt'),
    test_suite='run_tests.run',
    include_package_data=True,
)



from datetime import datetime

from django.db import models
from entity.models import Entity
from entity_event.models import Event
import uuid


class EmailManager(models.Manager):
    """
    Provides the ability to easily create emails with the recipients.
    """
    def create_email(self, recipients=None, **kwargs):
        """
        Note that we pop the scheduled time from the kwargs before creating the email
        and then update the scheduled time after the recipients have been added. This
        avoids the potential race condition of the email being created before it is
        picked up by a task that sends it.
        """
        scheduled = kwargs.pop('scheduled', datetime.utcnow())
        email = Email.objects.create(scheduled=None, **kwargs)
        if recipients:
            email.recipients.add(*recipients)

        email.scheduled = scheduled
        email.save()
        return email


class Email(models.Model):
    """Save an Email object and it is sent automagically!

    If an Email object is saved without a `scheduled` time, it will
    immediately and automatically be sent.

    If an Email object is saved with a `scheduled` time, it will be
    sent at that time.

    Emails can be sent to an individual, by setting subentities to
    False, or to a group, by setting send_to to a superentity, like a
    Team or Organization, and setting subentities to True. This applies
    to every entity in the recipients list.

    Sending an email happens automatically, and consists of rendering
    the given template with the given context.

    Emails will not be sent to any individual who has unsubscribed
    from emails of that source.

    Emails are viewable online and identified with their view_uid UUID
    """
    view_uid = models.UUIDField(default=uuid.uuid4, editable=False)
    event = models.ForeignKey(Event, on_delete=models.CASCADE)
    recipients = models.ManyToManyField(Entity)
    subject = models.CharField(max_length=256)
    from_address = models.CharField(max_length=256, default='')
    uid = models.CharField(max_length=100, unique=True, null=True, default=None)
    # The `scheduled` field uses a default value of the datetime.utcnow function.
    # This means it will be called any time a new Email is created, but it also
    # allows for a different schedule to be set (to schedule the email for some
    # time in the future), which would not be possible with an auto_add_now=True.
    scheduled = models.DateTimeField(null=True, default=datetime.utcnow)

    # The time that the email was actually sent, or None if the email is still waiting to be sent
    sent = models.DateTimeField(null=True, default=None)

    # Any exception that occurred when attempting to send the email last
    exception = models.TextField(default=None, null=True)

    objects = EmailManager()

    def render(self, medium):
        """
        Renders the event, assuming it has already had its context and renderers prefetched.
        """
        self.event.context['entity_emailer_id'] = str(self.view_uid)
        return self.event.render(medium)
